from utils import get_ms_time


class Block(object):
    _block_id = 1

    def __init__(self,
                 priority=0,
                 block_id=-1,
                 bytes_size=200000,
                 deadline=0.2,
                 timestamp=None):

        self.priority = priority
        self.block_id = block_id if block_id != -1 else Block.get_next_block_id()
        self.size = bytes_size
        self.deadline = deadline
        self.timestamp = timestamp if not timestamp is None else get_ms_time(1)
        # emulator params
        self.send_delay = 0
        self.latency = 0

        # log params
        self.finish_timestamp = -1
        self.miss_ddl = 0
        self.split_nums = -1
        self.finished_bytes = 0

    @classmethod
    def get_next_block_id(cls):
        ret = cls._block_id
        cls._block_id += 1
        return ret

    def get_cost_time(self):

        return self.send_delay + self.latency

    def get_block_info(self):
        return {
            "Block_id" : self.block_id,
            "Priority" : self.priority,
            "Deadline" : self.deadline,
            "Create_time" : self.timestamp,
            "Size" : self.size
        }

    def is_miss_ddl(self):
        if self.finish_timestamp == -1:
            return False
        return self.finish_timestamp-self.timestamp > self.deadline

    def __str__(self):

        return str(self.__dict__)

    def trans2dict(self):
        ret = self.get_block_info()
        ret.update({
            "Send_delay" : self.send_delay,
            "Latency" : self.latency,
            "Finish_timestamp" : self.finish_timestamp,
            "Miss_ddl" : self.miss_ddl,
            "Split_nums" : self.split_nums,
            "Finished_bytes" : self.finished_bytes
        })
        return ret



from config.constant import *
from utils import debug_print
from objects.cc_base import CongestionControl
import numpy as np;
import tensorflow as tf
import random

np.random.seed(2)
tf.set_random_seed(2)  # reproducible

# Superparameters
OUTPUT_GRAPH = False
MAX_EPISODE = 3000
DISPLAY_REWARD_THRESHOLD = 200  # renders environment if total episode reward is greater then this threshold
MAX_EP_STEPS = 1000   # maximum time step in one episode
RENDER = False  # rendering wastes time
GAMMA = 0.9     # reward discount in TD error
LR_A = 0.001    # learning rate for actor
LR_C = 0.01     # learning rate for critic

class Actor(object):
    def __init__(self, sess, n_features, n_actions, lr=0.001):
        self.sess = sess

        self.s = tf.placeholder(tf.float32, [1, n_features], "state")
        self.a = tf.placeholder(tf.int32, None, "act")
        self.td_error = tf.placeholder(tf.float32, None, "td_error")  # TD_error

        with tf.variable_scope('Actor'):
            l1 = tf.layers.dense(
                inputs=self.s,
                units=20,    # number of hidden units
                activation=tf.nn.relu,
                kernel_initializer=tf.random_normal_initializer(0., .1),    # weights
                bias_initializer=tf.constant_initializer(0.1),  # biases
                name='l1'
            )

            self.acts_prob = tf.layers.dense(
                inputs=l1,
                units=n_actions,    # output units
                activation=tf.nn.softmax,   # get action probabilities
                kernel_initializer=tf.random_normal_initializer(0., .1),  # weights
                bias_initializer=tf.constant_initializer(0.1),  # biases
                name='acts_prob'
            )

        with tf.variable_scope('exp_v'):
            log_prob = tf.log(self.acts_prob[0, self.a])
            self.exp_v = tf.reduce_mean(log_prob * self.td_error)  # advantage (TD_error) guided loss

        with tf.variable_scope('train'):
            self.train_op = tf.train.AdamOptimizer(lr).minimize(-self.exp_v)  # minimize(-exp_v) = maximize(exp_v)

    def learn(self, s, a, td):
        s = s[np.newaxis, :]
        feed_dict = {self.s: s, self.a: a, self.td_error: td}
        _, exp_v = self.sess.run([self.train_op, self.exp_v], feed_dict)
        return exp_v

    def choose_action(self, s):
        s = s[np.newaxis, :]
        probs = self.sess.run(self.acts_prob, {self.s: s})   # get probabilities for all actions
        temp_p = probs.ravel()
        print(temp_p)
        if np.isnan(temp_p[0]):
            temp_p[0] = random.uniform(0, 1.0)
            temp_p[1] = random.uniform(0,1-temp_p[0])
            temp_p[2] = 1 - temp_p[0] - temp_p[1]
        return np.random.choice(np.arange(probs.shape[1]), p=temp_p)   # return a int

class Critic(object):
    def __init__(self, sess, n_features, lr=0.01):
        self.sess = sess

        self.s = tf.placeholder(tf.float32, [1, n_features], "state")
        self.v_ = tf.placeholder(tf.float32, [1, 1], "v_next")
        self.r = tf.placeholder(tf.float32, None, 'r')

        with tf.variable_scope('Critic'):
            l1 = tf.layers.dense(
                inputs=self.s,
                units=20,  # number of hidden units
                activation=tf.nn.relu,  # None
                # have to be linear to make sure the convergence of actor.
                # But linear approximator seems hardly learns the correct Q.
                kernel_initializer=tf.random_normal_initializer(0., .1),  # weights
                bias_initializer=tf.constant_initializer(0.1),  # biases
                name='l1'
            )

            self.v = tf.layers.dense(
                inputs=l1,
                units=1,  # output units
                activation=None,
                kernel_initializer=tf.random_normal_initializer(0., .1),  # weights
                bias_initializer=tf.constant_initializer(0.1),  # biases
                name='V'
            )

        with tf.variable_scope('squared_TD_error'):
            self.td_error = self.r + GAMMA * self.v_ - self.v
            self.loss = tf.square(self.td_error)    # TD_error = (r+gamma*V_next) - V_eval
        with tf.variable_scope('train'):
            self.train_op = tf.train.AdamOptimizer(lr).minimize(self.loss)

    def learn(self, s, r, s_):
        s, s_ = s[np.newaxis, :], s_[np.newaxis, :]

        v_ = self.sess.run(self.v, {self.s: s_})
        td_error, _ = self.sess.run([self.td_error, self.train_op],
                                          {self.s: s, self.v_: v_, self.r: r})
        return td_error

EPISODE = 20 # change every EPISODE times
N_F = EPISODE * 3 # speed,losepacket,application_speed
N_A = 3 # +10,0,-10
Lambda_init = 1.0 # random choose
random_counter_init = 100 # decline Lambda after random_counter
MAX_BANDWITH = 15000 # standardlize to 1

sess = tf.Session()

actor = Actor(sess, n_features=N_F, n_actions=N_A, lr=LR_A)
critic = Critic(sess, n_features=N_F, lr=LR_C)     # we need a good teacher, so the teacher should learn faster than the actor

sess.run(tf.global_variables_initializer())

class RL(CongestionControl):

    def __init__(self):
        super(RL, self).__init__()
        self.USE_CWND = False
        self.send_rate = 40.0
        self.cwnd = 5000

        self.counter = 0 # EPISODE counter

        self.result_list = []
        self.last_state = []
        self.Lambda = Lambda_init
        self.random_counter = random_counter_init
        
        for i in range(EPISODE):
            self.last_state.append(40.0/MAX_BANDWITH)
        for i in range(EPISODE):
            self.last_state.append(0)
        for i in range(EPISODE):
            self.last_state.append(40.0/MAX_BANDWITH)

    def cc_trigger(self, data):

        packet_type = data["event_type"]
        event_time = data["event_time"]

        if packet_type == PACKET_TYPE_DROP:
            self.result_list.append(1)
        else:
            self.result_list.append(0)

        self.counter += 1
        if self.counter == EPISODE: # choose action every EPISODE times
            self.counter = 0
            print()
            print("EPISODE:")
            print()


            # declining random rate
            self.random_counter-=1
            if self.random_counter <= 0:
                self.Lambda /=2.0
                if self.Lambda < 0.05:
                    self.Lambda = 0.05
                self.random_counter = 0

            # reward
            r = 0
            for i in range(EPISODE):
                if self.result_list[i] == 0:
                    r += self.send_rate
                else:
                    r += -self.send_rate * EPISODE

            # current_state
            s_ = []
            for i in range(EPISODE):
                s_.append(self.send_rate/MAX_BANDWITH)
            for i in range(EPISODE):
                s_.append(self.result_list[i])
            for i in range(EPISODE):
                s_.append(self.send_rate/MAX_BANDWITH)
            s_array=np.array(s_)

            # choose action and explore
            a = actor.choose_action(s_array)
            if random.random() < self.Lambda:
                a = random.randint(0,2)
            print("action:",a)

            if a == 0:
                self.send_rate += 10.0
            elif a == 1:
                self.send_rate += 0.0
            else:
                self.send_rate += -10.0
                if self.send_rate < 40.0:
                    self.send_rate = 40.0

            # last state
            s = np.array(self.last_state)

            td_error = critic.learn(s, r, s_array)

            # debug
            sums = 0
            for i in range(EPISODE,2*EPISODE):
                sums += self.last_state[i]
            print("last_state:",self.last_state[0] * MAX_BANDWITH,sums)
            sums = 0
            for i in range(EPISODE,2*EPISODE):
                sums += s_[i]
            print("present_state:",s_[0] * MAX_BANDWITH,sums)
            print("r",r)
            print("td_error",td_error)

            if self.last_state[0] == self.send_rate:
                a = 1
            elif self.last_state[0] > self.send_rate:
                a = 2
            else:
                a = 0

            actor.learn(s, a, td_error)

            self.last_state = s_
            self.result_list = []

    def append_input(self, data):
        self._input_list.append(data)

        if data["event_type"] != PACKET_TYPE_TEMP:
            self.cc_trigger(data)
            return {
                "cwnd" : self.cwnd,
                "send_rate" : self.send_rate
            }
        return None


#!/usr/bin/python
# -*- coding: utf-8 -*-

'''
# @ModuleName : double_flow
# @Function :
# @Author : azson
# @Time : 2020/4/10 15:52
'''

from objects.pcc_emulator import PccEmulator
from utils import analyze_pcc_emulator, plot_cwnd, plot_rate
import os, sys, inspect, random
from config.constant import *
from objects.windows_based_sender import Sender as WinSender
from objects.engine import Engine
from objects.link import Link

from player.examples.reno import Reno
# from player.examples.simple_bbr import BBR
from player.examples.match_trace_rate import MTR
from player.packet_selection import Solution as PacketSelection
from objects.cc_base import CongestionControl


currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(currentdir)
sys.path.insert(0, parentdir)


# class BbrSolution(BBR, PacketSelection):
#     pass


class RenoSolution(Reno, PacketSelection):
    pass


class NormalSolution(MTR, PacketSelection):
    pass
    # def __init__(self):
    #     super().__init__()
    #     self.send_rate = 100
    #     self.USE_CWND = False
    #
    # def make_decision(self, cur_time):
    #     return {
    #         "send_rate": self.send_rate
    #     }


def create_2flow_emulator(solution, block_file=None, trace_file=None, **kwargs):

    emulator = PccEmulator(
        block_file=block_file,
        trace_file=trace_file,
        senders=[],
        links=[],
        **kwargs
    )
    emulator.trace_list = emulator.get_trace()
    queue = int(random.uniform(*emulator.queue_range))
    emulator.links = [Link(emulator.trace_list, queue), Link([], queue)]

    solution_1 = solution
    sender_1 = WinSender(emulator.links, 0, emulator.features, history_len=emulator.history_len, solution=solution_1)
    sender_1.init_application(emulator.block_file)

    solution_2 = NormalSolution()
    solution_2.init_trace(emulator.trace_file)
    sender_2 = WinSender(emulator.links, 0, emulator.features, history_len=emulator.history_len, solution=solution_2)
    # sender_2.init_application(emulator.block_file, ENABLE_BLOCK_LOG=False)

    emulator.senders = [sender_1, sender_2]
    emulator.net = Engine(emulator.senders, emulator.links)

    return emulator


if __name__ == '__main__':

    block_file = "config/block.txt"
    trace_file = "config/trace.txt"
    log_file = "output/pcc_emulator.log"
    log_packet_file = "output/packet_log/packet-0.log"

    new_trace_file = "scripts/first_group/traces_81.txt"
    new_block_files = ["config/data_video.csv", "config/data_audio.csv"]

    emulator = PccEmulator(
        block_file=block_file,
        trace_file=trace_file,
        senders=[],
        links=[]
    )
    emulator.trace_list = emulator.get_trace()
    queue = int(random.uniform(*emulator.queue_range))
    emulator.links = [Link(emulator.trace_list, queue), Link([], queue)]

    solution_2 = RenoSolution()
    # solution_2.init_trace(emulator.trace_file)
    sender_2 = WinSender(emulator.links, 0, emulator.features, history_len=emulator.history_len, solution=solution_2)
    # sender_2.init_application(emulator.block_file, ENABLE_BLOCK_LOG=False)

    emulator.senders = [sender_2]
    emulator.net = Engine(emulator.senders, emulator.links)

    print(emulator.run_for_dur(15))
    # emulator.dump_events_to_file(log_file)
    emulator.print_debug()
    # from qoe_model import cal_qoe
    # print(cal_qoe(0.9))
    analyze_pcc_emulator(log_packet_file, file_range="all")
    plot_cwnd(log_packet_file, trace_file=trace_file, file_range="all")
    # plot_throughput(log_packet_file, trace_file=trace_file, file_range="all")


from config.constant import *
from utils import debug_print
from objects.cc_base import CongestionControl


class Reno(CongestionControl):

    def __init__(self):
        super(Reno, self).__init__()
        self.ssthresh = float("inf")
        self.curr_state = "slow_start"
        self.states = ["slow_start", "congestion_avoidance", "fast_recovery"]
        self.drop_nums = 0
        self.ack_nums = 0

        self.cur_time = -1
        self.last_cwnd = 0
        self.instant_drop_nums = 0

    def cc_trigger(self, data):
        packet_type = data["event_type"]
        event_time = data["event_time"]

        if self.cur_time < event_time:
            self.last_cwnd = 0
            self.instant_drop_nums = 0

        if packet_type == PACKET_TYPE_DROP:
            if self.instant_drop_nums > 0:
                return
            self.instant_drop_nums += 1
            self.curr_state = self.states[2]
            self.drop_nums += 1
            self.ack_nums = 0
            # Ref 1 : For ensuring the event type, drop or ack?
            self.cur_time = event_time
            if self.last_cwnd > 0 and self.last_cwnd != self.cwnd:
                self.cwnd = self.last_cwnd
                self.last_cwnd = 0

        elif packet_type == PACKET_TYPE_FINISHED:
            # Ref 1
            if event_time <= self.cur_time:
                return
            self.cur_time = event_time
            self.last_cwnd = self.cwnd

            self.ack_nums += 1
            if self.curr_state == self.states[0]:
                if self.ack_nums == self.cwnd:
                    self.cwnd *= 2
                    self.ack_nums = 0
                if self.cwnd >= self.ssthresh:
                    self.curr_state = self.states[1]

            elif self.curr_state == self.states[1]:
                if self.ack_nums == self.cwnd:
                    self.cwnd += 1
                    self.ack_nums = 0

        if self.curr_state == self.states[2]:
            self.ssthresh = max(self.cwnd // 2, 1)
            self.cwnd = self.ssthresh
            self.curr_state = self.states[1]
        if self.drop_nums == 0:
            debug_print("Drop nums, Ack_nums")
            debug_print(self.drop_nums, self.ack_nums)

    def append_input(self, data):
        self._input_list.append(data)

        if data["event_type"] != PACKET_TYPE_TEMP:
            self.cc_trigger(data)
            return {
                "cwnd" : self.cwnd,
                "send_rate" : self.send_rate
            }
        return None




import os, sys, inspect, platform

currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(currentdir)
sys.path.insert(0, parentdir)
sys.path.insert(0, parentdir+"/simple_emulator")
# print(sys.path)

from objects.pcc_emulator import PccEmulator
from objects.cc_base import CongestionControl
from player.packet_selection import Solution as Packet_selection
from utils import *
from config.constant import *
from config import constant

from player.examples.reno import Reno
# from player.examples.simple_bbr import BBR
# from player.examples.RL import RL
from double_flow import create_2flow_emulator
from qoe_model import cal_qoe

# from scripts.block_trace_generator import generate_block_trace
# from scripts.network import create_network, create_trace


__all__ = ["PccEmulator", "CongestionControl", "Packet_selection", \
           "analyze_pcc_emulator", "plot_cwnd", "plot_rate", \
           "Reno", "create_2flow_emulator", "constant", \
           "cal_qoe"]

block_file = parentdir+"/simple_emulator"+"/config/block.txt"
trace_file = parentdir+"/simple_emulator"+"/config/trace.txt"
log_file = "output/pcc_emulator.log"
log_packet_file = "output/packet_log/packet-0.log"

new_trace_file = parentdir+"/simple_emulator"+"scripts/first_group/traces_1.txt"
new_block_files = [parentdir+"/simple_emulator"+"config/data_video.csv", parentdir+"/simple_emulator"+"config/data_audio.csv"]

try:
    if platform.system() == "Windows":
        # for windows
        os.system("rmdir /Q /S output")
    else:
        # for linux
        os.system("rm -rf output")

    # os.rmdir("output")
    os.mkdir("output")
    os.mkdir("output/packet_log")

except Exception as e:
    pass

# emulator = PccEmulator(
#     block_file=block_file,
#     trace_file=trace_file,
#     queue_range=(MIN_QUEUE, MAX_QUEUE)
# )



import sys, numpy, random
from sympy import Symbol, solve



def create_trace(change_scale=5, unchanged=4, cov=0.2, time_length=50, min_bitrate = 0.2, max_bitrate = 2):
    '''
    change_scale : the scale of probability to change state
    unchanged : the number of unchanged state
    cov : control the level of bitrate's similarity
    time_length : the numbers of bitrate
    '''

    # # get bitrate levels (in Mbps)
    # min_bitrate = 0.2
    # max_bitrate = 2
    steps = 10
    bitrate_states_low_var = []
    curr = min_bitrate
    for x in range(0, steps):
        bitrate_states_low_var.append(curr)
        curr += ((max_bitrate - min_bitrate) / (steps - 1))
        x += 1
    # list of transition probabilities
    transition_probabilities = []
    # assume you can go steps-1 states away (we will normalize this to the actual scenario)
    eq = -1
    x = Symbol("x", positive=True)
    for y in range(1, steps - 1):
        eq += (1 / x ** y)
    res = solve(eq)
    switch_parameter = res[0]

    for z in range(1, steps - 1):
        transition_probabilities.append(1 / switch_parameter * y)

    # two variance levels
    sigma_low = 1.0
    sigma_high = 1.0

    # probability of switching variance levels
    variance_switch_prob = 0.2

    # probability to stay in same state
    prob_stay = 1 - 1 / change_scale

    # takes a state and decides what the next state is
    def transition(state, variance):
        transition_prob = random.uniform(0, 1)

        # pick next variance first
        variance_switch = random.uniform(0, 1)
        next_variance = variance
        if (variance_switch < variance_switch_prob):
            if (next_variance == sigma_low):
                next_variance = sigma_high
            else:
                next_variance = sigma_low

        if transition_prob < prob_stay:  # stay in current state
            return (state, next_variance)
        else:  # pick appropriate state!
            curr_pos = state
            # first find max distance that you can be from current state
            max_distance = max(curr_pos, len(bitrate_states_low_var) - 1 - curr_pos)
            # cut the transition probabilities to only have possible number of steps
            curr_transition_probabilities = transition_probabilities[0:max_distance]
            trans_sum = sum(curr_transition_probabilities)
            normalized_transitions = [x / trans_sum for x in curr_transition_probabilities]
            # generate a random number and see which bin it falls in to
            trans_switch_val = random.uniform(0, 1)
            running_sum = 0
            num_switches = -1
            for ind in range(0, len(normalized_transitions)):
                if (trans_switch_val <= (normalized_transitions[ind] + running_sum)):  # this is the val
                    num_switches = ind
                    break
                else:
                    running_sum += normalized_transitions[ind]

            # now check if there are multiple ways to move this many states away
            switch_up = curr_pos + num_switches
            switch_down = curr_pos - num_switches
            if (switch_down >= 0 and switch_up <= (len(bitrate_states_low_var) - 1)):  # can go either way
                x = random.uniform(0, 1)
                if (x < 0.5):
                    return (switch_up, next_variance)
                else:
                    return (switch_down, next_variance)
            elif switch_down >= 0:  # switch down
                return (switch_down, next_variance)
            else:  # switch up
                return (switch_up, next_variance)

    current_state = random.randint(0, len(bitrate_states_low_var) - 1)
    current_variance = cov * bitrate_states_low_var[current_state]
    time = 0
    cnt = 0
    bitrates = []
    while time < time_length:
        # prints timestamp (in seconds) and throughput (in Mbits/s)
        if cnt <= 0:
            noise = numpy.random.normal(0, current_variance, 1)[0]
            cnt = unchanged
        gaus_val = max(0.1, bitrate_states_low_var[current_state] + noise)
        cnt -= 1
        print(str(time) + " " + str(gaus_val))
        bitrates.append(gaus_val)
        next_vals = transition(current_state, current_variance)
        if current_state != next_vals[0]:
            cnt = 0
        current_state = next_vals[0]
        current_variance = cov * bitrate_states_low_var[current_state]
        time += 1
    return bitrates

def create_network(row, length, min_bitrate, max_bitrate):
    for idx in range(50, 50 + length, 2):
        trace_list = []
        cs = random.randint(3, 10)
        uc = random.randint(3, 6)
        cv = random.uniform(0, 1)
        bw = create_trace(change_scale=cs, unchanged=uc, cov=cv, time_length=row, min_bitrate = min_bitrate, max_bitrate = max_bitrate)
        for i in range(row):
            trace_list.append([i , bw.pop(), 0, 0.001])
        with open("first_group/traces_"+str(idx + 1)+".txt", "w+") as f:
            for i in range(len(trace_list)):
                for j in range(3):
                    f.write(str(trace_list[i][j]) + ',')
                f.write(str(trace_list[i][3]) + '\n')


if __name__ == "__main__":
    create_network(50, 50, 1, 4)



#!/usr/bin/python
# -*- coding: utf-8 -*-

'''
# @ModuleName : main
# @Function : 
# @Author : azson
# @Time : 2020/3/2 20:02
'''

from objects.pcc_emulator import PccEmulator
from utils import analyze_pcc_emulator, plot_cwnd, plot_rate
import os, sys, inspect
from config.constant import *

currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(currentdir)
sys.path.insert(0, parentdir)

from player.aitrans_solution import Solution as s1
from player.aitrans_solution2 import Solution as s2
from player.aitrans_3 import Solution as s3


if __name__ == '__main__':

    block_file = "config/block.txt"
    trace_file = "config/trace.txt"
    log_file = "output/pcc_emulator.log"
    log_packet_file = "output/packet_log/packet-0.log"

    new_trace_file = "scripts/first_group/traces_1.txt"
    new_block_files = ["config/data_video.csv", "config/data_audio.csv"]

    emulator = PccEmulator(
        block_file=block_file,
        trace_file=trace_file,
        queue_range=(MIN_QUEUE, MAX_QUEUE)
    )

    print(emulator.run_for_dur(15))
    emulator.dump_events_to_file(log_file)
    emulator.print_debug()
    print(emulator.senders[0].rtt_samples)
    print(emulator.senders[0].application.ack_blocks)
    analyze_pcc_emulator(log_packet_file, file_range="all")
    plot_cwnd(log_packet_file, trace_file=trace_file, file_range="all")
    # plot_throughput(log_packet_file, trace_file=trace_file, file_range="all")


#!/usr/bin/python
# -*- coding: utf-8 -*-


from objects.pcc_emulator import PccEmulator
import os, sys, inspect
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(currentdir)
sys.path.insert(0, parentdir)
from config import constant

from qoe_difference import cal_distance, plt_qoe

if __name__ == '__main__':

    block_file = "config/block.txt"
    trace_file = "config/trace.txt"
    new_trace_file = "scripts/first_group/traces_51.txt"
    log_file = "output/pcc_emulator.log"
    log_packet_file = "output/packet_log/packet-0.log"
    pic = "qoemodel/qoes_buffer.png"
    idx, size = "queue_size", 50
    x = 0.9
    reno_arr = []
    bbr_arr = []
    for j in range(10, 60):
        constant.MAX_QUEUE = constant.MIN_QUEUE = j
        qoe_difference = cal_distance(block_file, trace_file, x)
        reno_arr.append(qoe_difference[0])
        bbr_arr.append(qoe_difference[1])

    plt_qoe(reno_arr, bbr_arr, pic, idx, size)

    with open("qoemodel/qoes_buffer.log","w+") as f:
        strs = ["buffer: MAX_QUEUE = 10, MIN_QUEUE = 59\n",
                "qoe = 0.9 * priority + 0.1 * deadline\n",
                str(reno_arr) + "\n",
                str(bbr_arr)]
        f.writelines(strs)










#!/usr/bin/python
# -*- coding: utf-8 -*-

'''
# @ModuleName : block_trace_generator
# @Function : 
# @Author : azson
# @Time : 2020/4/23 9:18
'''
import random


def generate_block_trace(network_trace, avg_block_size=200000, dur_sec=15, use_noisy=False, output_file="data_block.csv"):
    trace_list = []
    block_list = []
    with open(network_trace, "r") as f:
        for line in f.readlines():
            trace_list.append(list(
                map(lambda x: float(x), line.split(","))
            ))
    cur_time = 0
    for idx, item in enumerate(trace_list):
        time, bw, loss_rate, p_time = item
        bw *= 10**6
        if idx == 0:
            continue
        end_time = trace_list[-1][0]+5 if idx == len(trace_list)-1 else trace_list[idx+1][0]
        end_time = min(end_time, dur_sec)
        while cur_time < end_time:
            new_block_create_time = cur_time
            new_block_size = avg_block_size
            if use_noisy:
                new_block_size = new_block_size * (0.9+0.2*random.random())

            block_list.append([new_block_create_time, new_block_size])
            cur_time += new_block_size / bw
        if cur_time >= dur_sec:
            break
    if output_file:
        with open(output_file, 'w') as f:
            for line in block_list:
                f.write(','.join(list(map(lambda x:str(x), line))))
                f.write('\n')

    return block_list



if __name__ == '__main__':
    # network_trace = "first_group/traces_1.txt"
    network_trace = "../config/trace.txt"
    block_trace = generate_block_trace(network_trace, use_noisy=True)
    print(block_trace)
    print(len(block_trace))


DELTA_SCALE = 0.9

MAX_CWND = 5000
MIN_CWND = 4

MAX_RATE = 1000
MIN_RATE = 40

REWARD_SCALE = 0.001

EVENT_TYPE_SEND = 'S'
EVENT_TYPE_ACK = 'A'

BYTES_PER_PACKET = 1500
HEAD_PER_PACKAET = 20

LATENCY_PENALTY = 1.0
LOSS_PENALTY = 1.0

USE_LATENCY_NOISE = False
MAX_LATENCY_NOISE = 1.1

USE_CWND = True
MAX_QUEUE = 55
MIN_QUEUE = 55

PACKET_TYPE_FINISHED = 'F'
PACKET_TYPE_TEMP = 'T'
PACKET_TYPE_DROP = 'D'

ENABLE_HASH_CHECK = False
ENABLE_DEBUG = False
ENABLE_LOG = True
MAX_PACKET_LOG_ROWS = 4000
ALERT_CIRCLE = 5


class CongestionControl(object):

    def __init__(self):
        self._input_list = []
        self.cwnd = 1
        self.send_rate = float("inf")
        self.pacing_rate = float("inf")
        self.call_nums = 0

    def make_decision(self, cur_time):
        """call this when sender send packet"""
        self.call_nums += 1

        output = {
            "cwnd" : self.cwnd,
            "send_rate" : self.send_rate,
            "extra" : { }
        }

        return output

    def append_input(self, data):
        self._input_list.append(data)

        return None


import json, random, os, inspect
import numpy as np

from utils import (
    get_emulator_info, measure_time
)
from objects.windows_based_sender import Sender as W_sender
from objects.link import Link
from objects.engine import Engine

from player.aitrans_solution import Solution as Aitrans_solution
from config import constant

currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(currentdir)
if "simple_emulator" not in parentdir:
    parentdir += "simple_emulator"


class PccEmulator(object):

    def __init__(self,
                 block_file=None,
                 trace_file=None,
                 queue_range=None,
                 solution=None,
                 **kwargs):
        # do configure on constant
        self.update_config(kwargs)

        self.trace_cols = ("time", "bandwith", "loss_rate", "delay")
        self.queue_range = queue_range if queue_range else (constant.MIN_QUEUE, constant.MAX_QUEUE)
        self.trace_file = trace_file if trace_file else parentdir + "/config/trace.txt"
        if isinstance(self.trace_file, int):
            self.trace_file = parentdir + "/scripts/first_group/traces_%d.txt" % self.trace_file
        self.block_file = block_file if block_file else parentdir + "/config/block.txt"
        self.event_record = { "Events" : [] }

        # unkown params
        self.features = [] # ["send rate", "recv rate"]
        self.history_len = 1
        self.steps_taken = 0

        self.links = None
        self.senders = None
        self.solution = solution
        self.trace_list = None
        if "senders" in kwargs:
            self.senders = kwargs["senders"]
        if "links" in kwargs:
            self.links = kwargs["links"]
        if "senders" not in kwargs and "links" not in kwargs:
            self.create_new_links_and_senders()
        self.net = Engine(self.senders, self.links)

    def update_config(self, extra):
        """make it available that change some variable in constant.py by the way of transmitting function parameters."""
        if "USE_CWND" in extra:
            constant.USE_CWND = extra["USE_CWND"]
        if "ENABLE_DEBUG" in extra:
            constant.ENABLE_DEBUG = extra["ENABLE_DEBUG"]
        if "ENABLE_LOG" in extra:
            constant.ENABLE_LOG = extra["ENABLE_LOG"]
        if "ENABLE_HASH_CHECK" in extra:
            constant.ENABLE_HASH_CHECK = extra["ENABLE_HASH_CHECK"]
        if "MAX_PACKET_LOG_ROWS" in extra:
            constant.ENABLE_LOG = extra["MAX_PACKET_LOG_ROWS"]
        if "MIN_QUEUE" in extra:
            constant.MIN_QUEUE = extra["MIN_QUEUE"]
        if "MAX_QUEUE" in extra:
            constant.MAX_QUEUE = extra["MAX_QUEUE"]

    def get_trace(self):
        """init the "trace_list" according to the trace file."""
        trace_list = []
        with open(self.trace_file, "r") as f:
            for line in f.readlines():
                trace_list.append(list(
                    map(lambda x: float(x), line.split(","))
                ))
                if len(trace_list[-1]) != len(self.trace_cols):
                    raise ValueError("Trace file error!\nPlease check its format like : {0}".format(self.trace_cols))

        if len(trace_list) == 0:
            raise ValueError("Trace file error!\nThere is no data in the file!")

        return trace_list

    def create_new_links_and_senders(self):
        """create links and senders in this network."""
        self.trace_list = self.get_trace()
        # queue = 1 + int(np.exp(random.uniform(*self.queue_range)))
        # print("queue size : %d" % queue)
        # bw = self.trace_list[0][1]

        queue = int(random.uniform(*self.queue_range))
        self.links = [Link(self.trace_list, queue) , Link([], queue)]
        #self.senders = [Sender(0.3 * bw, [self.links[0], self.links[1]], 0, self.history_len)]
        #self.senders = [Sender(random.uniform(0.2, 0.7) * bw, [self.links[0], self.links[1]], 0, self.history_len)]
        solution = Aitrans_solution() if not self.solution else self.solution
        # support change type of cc by aitrans solution
        if hasattr(solution, "USE_CWND"):
            constant.USE_CWND = solution.USE_CWND

        self.senders = [W_sender(self.links, 0, self.features,
                               history_len=self.history_len, solution=solution)]
        for item in self.senders:
            item.init_application(self.block_file)

    # @measure_time()
    def run_for_dur(self, during_time=float("inf")):
        """run this emulator for time of "dur_time"."""
        # action = [0.9, 0.9]
        # for i in range(len(self.senders)):
        #     self.senders[i].apply_rate_delta(action[0])
        #     if USE_CWND:
        #         self.senders[i].apply_cwnd_delta(action[1])

        reward = self.net.run_for_dur(during_time)
        for sender in self.senders:
            sender.record_run()

        sender_obs = self._get_all_sender_obs()
        sender_mi = self.senders[0].get_run_data()
        event = get_emulator_info(sender_mi)
        event["reward"] = reward
        self.event_record["Events"].append(event)
        if event["Latency"] > 0.0:
            self.run_dur = 0.5 * sender_mi.get("avg latency")

        return event, sender_obs

    def print_debug(self):
        print("---Link Debug---")
        for link in self.links:
            link.print_debug()
        print("---Sender Debug---")
        for sender in self.senders:
            sender.print_debug()

    def reset(self):
        self.steps_taken = 0
        self.net.reset()
        self.create_new_links_and_senders()
        self.net = Engine(self.senders, self.links)
        self.episodes_run += 1
        if self.episodes_run > 0 and self.episodes_run % 100 == 0:
            self.dump_events_to_file("pcc_env_log_run_%d.json" % self.episodes_run)
        self.event_record = {"Events": []}
        self.net.run_for_dur(self.run_dur)
        self.net.run_for_dur(self.run_dur)
        self.reward_ewma *= 0.99
        self.reward_ewma += 0.01 * self.reward_sum
        print("Reward: %0.2f, Ewma Reward: %0.2f" % (self.reward_sum, self.reward_ewma))
        self.reward_sum = 0.0
        return self._get_all_sender_obs()

    def close(self):
        if self.viewer:
            self.viewer.close()
            self.viewer = None

    def dump_events_to_file(self, filename):
        with open(filename, 'w') as f:
            json.dump(self.event_record, f, indent=4)

    def _get_all_sender_obs(self):
        sender_obs = self.senders[0].get_obs()
        sender_obs = np.array(sender_obs).reshape(-1, )
        # print(sender_obs)
        return sender_obs





from player.congestion_control_algorithm import Solution as Cc_solution
from player.packet_selection import Solution as Packet_solution
# from player.examples.simple_bbr import BBR

class Solution(Cc_solution, Packet_solution):
    pass



from utils import measure_time


class Solution(object):

    # @measure_time()
    def select_packet(self, cur_time, packet_queue):
        '''
        The alogrithm to select the packet which will be sended in next.
        The following example is selecting packet by the create time firstly, and radio of rest life time to deadline secondly.
        :param cur_time: float
        :param packet_queue: the list of Packet.You can get more detail about Block in objects/packet.py
        :return: int
        '''
        def is_better(packet):
            best_block_create_time = best_packet.block_info["Create_time"]
            packet_block_create_time = packet.block_info["Create_time"]
            # if packet is miss ddl
            if (cur_time - packet_block_create_time) >= packet.block_info["Deadline"]:
                return False
            if (cur_time - best_block_create_time) >= best_packet.block_info["Deadline"]:
                return True
            if best_block_create_time != packet_block_create_time:
                return best_block_create_time > packet_block_create_time
            return (cur_time - best_block_create_time) * best_packet.block_info["Deadline"] > \
                   (cur_time - packet_block_create_time) * packet.block_info["Deadline"]

        best_packet_idx = -1
        best_packet = None
        for idx, item in enumerate(packet_queue):
            if best_packet is None or is_better(item) :
                best_packet_idx = idx
                best_packet = item

        return best_packet_idx



from objects.sender import Sender as Super_sender
import heapq


class Sender(Super_sender):

    def slide_windows(self, cur_time, queue_size):
        ret = []
        for i in range(int(self.cwnd) - queue_size):
            if len(self.wait_for_push_packets) == 0:
                _packet = self.select_packet(cur_time + (1.0 / self.rate))
                if _packet is None:
                    return ret
            else:
                _packet = heapq.heappop(self.wait_for_push_packets)[2]
            ret.append(_packet)

        return ret



from player.congestion_control_algorithm import Solution as Cc_solution
from player.packet_selection import Solution as Packet_solution
from player.examples.reno import Reno

class Solution(Reno, Packet_solution):
    pass



#!/usr/bin/python
# -*- coding: utf-8 -*-


from objects.pcc_emulator import PccEmulator
import os, sys, inspect
from config.constant import *
import numpy as np
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(currentdir)
sys.path.insert(0, parentdir)
import matplotlib.pyplot as plt
import numpy as np
import json

from player.aitrans_solution import Solution as s1
from player.aitrans_solution2 import Solution as s2
from player.aitrans_3 import Solution as s3

from qoe_model import cal_qoe
from double_flow import create_2flow_emulator


def cal_distance_double(block_file, trace_file, x):

    emulator1 = create_2flow_emulator(s1(), block_file, trace_file)
    emulator1.run_for_dur(21)
    reno_qoe = cal_qoe(x)

    emulator2 = create_2flow_emulator(s2(), block_file, trace_file)
    emulator2.run_for_dur(21)
    bbr_qoe = cal_qoe(x)

    tmp = s3()
    tmp.init_trace(trace_file)
    emulator3 = create_2flow_emulator(tmp, block_file, trace_file)
    emulator3.run_for_dur(21)
    mtr_qoe = cal_qoe(x)

    return [reno_qoe, bbr_qoe, mtr_qoe]


def cal_distance_single(block_file, trace_file, x):
    emulator1 = PccEmulator(
        block_file=block_file,
        trace_file=trace_file,
        solution=s1(),
        USE_CWND=True
    )
    emulator1.run_for_dur()
    reno_qoe = cal_qoe(x)

    emulator2 = PccEmulator(
        block_file=block_file,
        trace_file=trace_file,
        solution=s2(),
        USE_CWND=True
    )
    emulator2.run_for_dur()
    bbr_qoe = cal_qoe(x)

    tmp = s3()
    tmp.init_trace(trace_file)
    emulator3 = PccEmulator(
        block_file=block_file,
        trace_file=trace_file,
        solution=tmp,
        USE_CWND=False
    )
    emulator3.run_for_dur()
    mtr_qoe = cal_qoe(x)
    return [reno_qoe, bbr_qoe, mtr_qoe]


def plt_qoe(reno_arr, bbr_arr, pic, xidx, size, mtr_arr=None):
    x = np.linspace(1, 100, size)
    fig, ax = plt.subplots()
    ax.plot(x,reno_arr, color="blue", label="reno_qoe")
    ax.plot(x, bbr_arr, color="red", label="bbr_qoe")
    if mtr_arr:
        ax.plot(x, mtr_arr, color="green", label="mtr_qoe")
    ax.set_xlabel(xidx)
    ax.set_ylabel("qoe")
    ax.legend()
    plt.savefig(pic)


if __name__ == '__main__':

    block_file = "config/block.txt"
    log_file = "output/pcc_emulator.log"
    log_packet_file = "output/packet_log/packet-0.log"
    pic = "qoemodel/qoe_difference.png"
    idx,size = "trace_index",10
    x = 0.9
    reno_arr = []
    bbr_arr = []
    mtr_arr = []
    for j in range(1, 101, 10):
        trace_file = "scripts/first_group/traces_" + str(j) + ".txt"
        qoe_difference = cal_distance_double(block_file, trace_file, x)
        reno_arr.append(qoe_difference[0])
        bbr_arr.append(qoe_difference[1])
        mtr_arr.append(qoe_difference[2])

    plt_qoe(reno_arr, bbr_arr, pic, idx, size, mtr_arr)
    MIN_QUEUE = MAX_QUEUE = queue_size = 55
    with open("qoemodel/qoe_difference.log","w+") as f:
        strs = ["trace numbers : 50\n", "buffer: MAX_QUEUE = MIN_QUEUE = "+
                str(queue_size) + "\n",
                "bw : 0.1 ~ 2 MB \n",
                "qoe = " + str(x) + " * priority + "+ str(1 - x) + " * deadline\n",
                str(reno_arr) + "\n",
                str(bbr_arr)]
        f.writelines(strs)










import heapq, random, json
from config.constant import *
from utils import get_packet_type, get_emulator_info, debug_print
from config import constant


class Engine():

    def __init__(self, senders, links):
        self.q = []
        self.cur_time = 0.0
        self.senders = senders
        self.links = links
        self.queue_initial_packets()

        self.fir_log = True
        self.log_packet_file = "output/packet_log/packet-0.log"
        self.log_items = 0
        self.last_alert_time = 0

    def queue_initial_packets(self):
        """initial the packet queue in network"""
        for sender in self.senders:
            sender.register_network(self)
            sender.reset_obs()
            packet = sender.select_packet(1.0 / sender.rate) # sender.new_packet(1.0 / sender.rate)
            if packet:
                sender.in_event_nums += 1
                heapq.heappush(self.q, (1.0 / sender.rate, sender, packet))

    def reset(self):
        self.cur_time = 0.0
        self.q = []
        [link.reset() for link in self.links]
        [sender.reset() for sender in self.senders]
        self.queue_initial_packets()

    def get_cur_time(self):
        return self.cur_time

    def run_for_dur(self, dur):
        """run the system for time of "dur". """
        end_time = self.cur_time + dur
        for sender in self.senders:
            sender.reset_obs()

        while self.cur_time < end_time:
            if len(self.q) == 0:
                print("Time {}s : There is no packet from application~".format(self.cur_time))
                return

            event_time, sender, packet = heapq.heappop(self.q)
            self.log_packet(event_time, sender, packet)

            event_type, next_hop, cur_latency, dropped, packet_id = packet.parse()
            # print("Got event %s, to link %d, latency %f at time %f" % (event_type, next_hop, cur_latency, event_time))
            self.cur_time = event_time
            new_event_time = event_time
            new_event_type = event_type
            new_next_hop = next_hop
            new_latency = cur_latency
            new_dropped = dropped
            push_new_event = False

            if event_type == EVENT_TYPE_ACK:
                # got ack in source or destination
                if next_hop == len(sender.path):
                    packet.finish_time = event_time
                    self.append_cc_input(event_time, sender, packet)
                    sender.in_event_nums -= 1
                    if dropped:
                        sender.on_packet_lost(event_time, packet)
                        # print("Packet lost at time %f" % self.cur_time)
                    else:
                        # may acked packet which is not in window after packet loss
                        sender.on_packet_acked(packet.get_rtt(), packet)
                        # print("Packet acked at time %f" % self.cur_time)
                    # for windows-based cc
                    if sender.USE_CWND:
                        # continue ack may use same inflight numbers which will be limited to cwnd but redundancy in log
                        for _packet in sender.slide_windows(self.cur_time, sender.in_event_nums):
                            sender.in_event_nums += 1
                            # print("Now {}, packet id {}, create time {}".format(self.cur_time, _packet.packet_id, _packet.create_time))
                            # print("wait for push set {}".format(set([item[2].create_time for item in sender.wait_for_push_packets])))
                            heapq.heappush(self.q, (max(self.cur_time+(1.0 / sender.rate), _packet.create_time), \
                                                sender, _packet))

                # ack back to source
                else:
                    new_next_hop = next_hop + 1
                    link_latency = sender.path[next_hop].get_cur_latency(self.cur_time)
                    if USE_LATENCY_NOISE:
                        link_latency *= random.uniform(1.0, MAX_LATENCY_NOISE)
                    new_latency += link_latency
                    new_event_time += link_latency
                    push_new_event = True
            if event_type == EVENT_TYPE_SEND:
                if next_hop == 0:
                    # print("Packet sent at time %f" % self.cur_time)
                    if sender.can_send_packet(new_event_time):
                        pacing_time, extra_info = sender.on_packet_sent(new_event_time)
                        packet.pacing_delay += pacing_time
                        packet.extra = extra_info
                        push_new_event = True
                    else:
                        sender.in_event_nums -= 1
                        if sender.application:
                            sender.wait_for_push_packets.append([event_time, sender, packet])
                    # when do the packet create ? before or after pacing ?
                    _packet = sender.select_packet(new_event_time + (1.0 / sender.rate)) # new_packet(new_event_time + (1.0 / sender.rate))
                    if _packet:
                        heapq.heappush(sender.wait_for_push_packets, [_packet.create_time, sender, _packet])
                        if not sender.USE_CWND or int(sender.cwnd) > sender.in_event_nums:
                            item = heapq.heappop(sender.wait_for_push_packets)
                            sender.in_event_nums += 1
                            heapq.heappush(self.q, (max(new_event_time + (1.0 / sender.rate), item[0]), item[1], item[2]))
                    new_event_time += pacing_time
                else:
                    push_new_event = True

                if next_hop == sender.dest:
                    new_event_type = EVENT_TYPE_ACK
                new_next_hop = next_hop + 1
                link_latency = sender.path[next_hop].get_cur_latency(new_event_time)
                if USE_LATENCY_NOISE:
                    link_latency *= random.uniform(1.0, MAX_LATENCY_NOISE)
                new_latency += link_latency
                new_dropped = not sender.path[next_hop].packet_enters_link(new_event_time)
                new_event_time += link_latency

            if push_new_event:
                packet.next_hop = new_next_hop
                packet.packet_type = new_event_type
                packet.latency = new_latency
                packet.drop = new_dropped
                heapq.heappush(self.q, (new_event_time, sender, packet))
        self.close()
        sender_mi = self.senders[0].get_run_data()
        throughput = sender_mi.get("recv rate")
        latency = sender_mi.get("avg latency")
        loss = sender_mi.get("loss ratio")
        # bw_cutoff = self.links[0].bw * 0.8
        # lat_cutoff = 2.0 * self.links[0].dl * 1.5
        # loss_cutoff = 2.0 * self.links[0].lr * 1.5
        # print("thpt %f, bw %f" % (throughput, bw_cutoff))
        # reward = 0 if (loss > 0.1 or throughput < bw_cutoff or latency > lat_cutoff or loss > loss_cutoff) else 1 #

        # Super high throughput
        # reward = REWARD_SCALE * (20.0 * throughput / RATE_OBS_SCALE - 1e3 * latency / LAT_OBS_SCALE - 2e3 * loss)

        # Very high thpt
        reward = (10.0 * throughput / (8 * BYTES_PER_PACKET) - 1e3 * latency - 2e3 * loss)

        # High thpt
        # reward = REWARD_SCALE * (5.0 * throughput / RATE_OBS_SCALE - 1e3 * latency / LAT_OBS_SCALE - 2e3 * loss)

        # Low latency
        # reward = REWARD_SCALE * (2.0 * throughput / RATE_OBS_SCALE - 1e3 * latency / LAT_OBS_SCALE - 2e3 * loss)
        # if reward > 857:
        # print("Reward = %f, thpt = %f, lat = %f, loss = %f" % (reward, throughput, latency, loss))

        # reward = (throughput / RATE_OBS_SCALE) * np.exp(-1 * (LATENCY_PENALTY * latency / LAT_OBS_SCALE + LOSS_PENALTY * loss))
        return reward * REWARD_SCALE

    def log_packet(self, event_time, sender, packet):
        """
        log the event information.
        :param event_time: the time happened this event.
        :param sender: the sender sent this packet.
        :param packet: type Packet.
        """
        # only log the first sender when ENABLE_LOG=True
        if not constant.ENABLE_LOG:
            return packet

        def get_true_log_file():
            """if the rows of single log file is limited to MAX_PACKET_LOG_ROWS, find the name of next new log file."""
            if isinstance(constant.MAX_PACKET_LOG_ROWS, int) and constant.MAX_PACKET_LOG_ROWS > 0:
                file_nums = self.log_items // constant.MAX_PACKET_LOG_ROWS
                if self.log_items and self.log_items % constant.MAX_PACKET_LOG_ROWS == 0:
                    self.fir_log = True
                return self.log_packet_file.replace('0', str(file_nums))
            return self.log_packet_file

        if constant.ENABLE_DEBUG and event_time - self.last_alert_time >= ALERT_CIRCLE:
            self.last_alert_time = event_time
            sender_mi = self.senders[0].get_run_data()
            print("Time : {}".format(event_time))
            print(get_emulator_info(sender_mi))

        if self.fir_log:
            self.fir_log = False
            with open(get_true_log_file(), "w") as f:
                pass

        log_data = {
            "Time" : event_time,
            "Waiting_for_ack_nums" : sender.get_waiting_ack_nums(),
            "Sender_id" : sender.id
        }
        log_data.update(packet.trans2dict())
        if sender.USE_CWND:
            log_data["Extra"]["Cwnd"] = sender.cwnd
        if sender.rate != float("inf"):
            log_data["Extra"]["Send_rate"] = sender.rate
        # log_data["Extra"]["in_event_nums"] = sender.in_event_nums
        # log_data["Extra"]["wait_for_select"] = len(sender.wait_for_select_packets)
        # log_data["Extra"]["wait_for_push"] = len(sender.wait_for_push_packets)
        with open(get_true_log_file(), "a") as f:
            f.write(json.dumps(log_data, ensure_ascii = False)+"\n")
        self.log_items += 1
        return packet

    def append_cc_input(self, event_time, sender, packet, event_type="packet"):
        """push the acked or lost packet event to it's sender"""
        # only use the solution's sender
        if sender.id != self.senders[0].id:
            return
        if event_type == "packet":
            data = {
                "event_time" : event_time,
                "event_type" : get_packet_type(sender, packet),
                "packet_information_dict" : packet.trans2dict()
            }
            data["packet_information_dict"]["Extra"]["inflight"] = sender.get_waiting_ack_nums()
        # todo : how to design system information
        elif event_type == "system":
            data = {

            }

        feed_back = sender.solution.append_input(data)
        if feed_back:
            sender.cwnd = feed_back["cwnd"] if "cwnd" in feed_back else sender.cwnd
            sender.rate = feed_back["send_rate"] if "send_rate" in feed_back else sender.rate
            sender.pacing_rate = feed_back["pacing_rate"] if "pacing_rate" in feed_back else sender.pacing_rate

    def close(self):
        """close all the application before closing this system."""

        for sender in self.senders:
            debug_print("sender {} wait_for_push_packets size {}".format(sender.id, len(sender.wait_for_push_packets)))
            if sender.application:
                sender.application.close()


import setuptools


with open("README.md", "r", encoding="utf8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="DTP_Emulator", # Replace with your own username
    version="1.1",
    author="azson",
    author_email="240326315@qq.com",
    description="DTP emulator",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Azson/DTP-emulator",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.0',
    package_data={
        '':["*txt", "*csv"]
    }
)



#!/usr/bin/python
# -*- coding: utf-8 -*-

'''
# @ModuleName : utils
# @Function : 
# @Author : azson
# @Time : 2020/1/8 15:59
'''

import time, json
from matplotlib import pyplot as plt
import numpy as np
from config.constant import *
from config import constant


def get_ms_time(rate=1000):

    return time.time()*rate


def measure_time():
    def wraps(func):
        def measure(*args,**kwargs):
            start = time.time()
            res = func(*args,**kwargs)
            end = time.time()
            if constant.ENABLE_DEBUG:
                with open("output/funtion_time_test.log", "a") as f:
                    f.write("function %s use time %s\n"%(func.__name__, (end-start)))
            return res
        return measure
    return wraps


def analyze_pcc_emulator(log_file, trace_file=None, rows=None, time_range=None, scatter=False, file_range=None, sender=None):

    plt_data = []
    if file_range:
        plt_data = compose_packet_logs(file_range)
    else:
        with open(log_file, 'r') as f:
            for line in f.readlines():
                plt_data.append(json.loads(line.replace("'", '"')))
    if not sender:
        sender = [1]
    plt_data = list(filter(lambda x:x["Type"]=='A' and x["Position"] == 2 and x["Sender_id"] in sender, plt_data))
    if time_range:
        plt_data = time_filter(plt_data, time_range)
    # priority by packet id
    # plt_data = sorted(plt_data, key=lambda x:int(x["Packet_id"]))[:rows]
    if isinstance(rows, int):
        plt_data = plt_data[:rows]

    pic_nums = 2
    font_size = 50
    tick_size = 50

    data_latency = []
    data_finish_time = []
    data_drop = []
    data_sum_time = []
    # data_miss_ddl = []
    for idx, item in enumerate(plt_data):
        if item["Type"] == 'A':
            if item["Drop"] == 1:
                data_drop.append(idx)
            else:
                data_latency.append(item["Latency"])
                data_finish_time.append(item["Time"])
                data_sum_time.append(item["Send_delay"] + item["Pacing_delay"] + item["Latency"])

            # if item["Block_info"]["Deadline"] < data_sum_time[-1]:
            #     data_miss_ddl.append(idx)

    pic = plt.figure(figsize=(50, 30 * pic_nums))
    # plot latency distribution
    ax = plt.subplot(pic_nums, 1, 1)
    ax.set_title("Acked packet latency distribution", fontsize=font_size)
    ax.set_ylabel("Latency / s", fontsize=font_size)
    ax.set_xlabel("Time / s", fontsize=font_size)
    if scatter:
        ax.scatter(data_finish_time, data_latency, label="Latency", s=200)
    else:
        ax.plot(data_finish_time, data_latency, label="Latency", linewidth=5)

    ax.scatter([plt_data[idx]["Time"] for idx in data_drop],
               [min(data_latency) / 2]*len(data_drop), label="Drop", s=300, c='r', marker='x')

    # plot average latency
    ax.plot([0, data_finish_time[-1] ], [np.mean(data_latency)]*2, label="Average Latency", linewidth=5,
            c='g')
    plt.legend(fontsize=font_size)
    ax.set_xlim(data_finish_time[0] / 2, data_finish_time[-1] * 1.2)
    plt.tick_params(labelsize=tick_size)

    # # plot miss deadline rate block
    # ax = plt.subplot(pic_nums, 1, 2)
    # ax.set_title("Acked packet lost distribution", fontsize=font_size)
    # ax.set_ylabel("Latency / s", fontsize=font_size)
    # ax.set_xlabel("Time / s", fontsize=font_size)
    # ax.scatter([data_finish_time[idx] for idx in data_drop],
    #                 [data_latency[idx] for idx in data_drop], label="Drop")
    # ax.scatter([data_finish_time[idx] for idx in data_miss_ddl],
    #                 [data_latency[idx] for idx in data_miss_ddl], label="Miss_deadline")
    # plt.legend(fontsize=font_size)
    # ax.set_xlim(data_finish_time[0] / 2, data_finish_time[-1] * 1.5)
    # plt.tick_params(labelsize=tick_size)

    # plot latency distribution
    ax = plt.subplot(pic_nums, 1, 2)
    ax.set_title("Acked packet RTT distribution", fontsize=font_size)
    ax.set_ylabel("RTT / s", fontsize=font_size)
    ax.set_xlabel("Time / s", fontsize=font_size)
    # ax.set_ylim(-np.min(data_sum_time)*2, np.max(data_sum_time)*2)

    ax.plot(data_finish_time, data_sum_time, label="Latency", linewidth=5)
    # plot average latency
    ax.plot([0, data_finish_time[-1]], [np.mean(data_sum_time)] * 2, label="Average Rtt",
            c='r', linewidth=5)

    # ax.set_xlim(data_finish_time[0]/2, data_finish_time[-1]*1.5)
    plt.tick_params(labelsize=tick_size)
    handles, labels = ax.get_legend_handles_labels()

    # plot bandwith
    if trace_file:
        tmp_ax = plot_trace(data_finish_time, ax, font_size, tick_size, trace_file)
        handles.extend(tmp_ax.get_legend_handles_labels()[0])
        labels.extend(tmp_ax.get_legend_handles_labels()[1])

    plt.legend(handles, labels, fontsize=font_size)

    plt.tight_layout()
    plt.legend(fontsize=font_size)
    plt.savefig("output/pcc_emulator-analysis.png")


def check_solution_format(input):

    if not isinstance(input, dict):
        raise TypeError("The return value should be a dict!")

    keys = ["cwnd", "send_rate"]
    if keys[0] in input or keys[1] in input:
        return input
    else:
        raise ValueError("One of the keys %s should in the return dict!" % (keys))


def get_emulator_info(sender_mi):

    event = {}
    event["Name"] = "Step"
    # event["Target Rate"] = sender_mi.target_rate
    event["Send Rate"] = sender_mi.get("send rate")
    event["Throughput"] = sender_mi.get("recv rate")
    event["Latency"] = sender_mi.get("avg latency")
    event["Loss Rate"] = sender_mi.get("loss ratio")
    event["Latency Inflation"] = sender_mi.get("sent latency inflation")
    event["Latency Ratio"] = sender_mi.get("latency ratio")
    event["Send Ratio"] = sender_mi.get("send ratio")
    # event["Cwnd"] = sender_mi.cwnd
    # event["Cwnd Used"] = sender_mi.cwnd_used

    return event


def analyze_application(acked_packets):
    pass


def get_packet_type(sender, packet):
    if packet.drop:
        return PACKET_TYPE_DROP
    if packet.packet_type == EVENT_TYPE_ACK and \
            packet.next_hop == len(sender.path):
        return PACKET_TYPE_FINISHED

    return PACKET_TYPE_TEMP


def debug_print(*args, **kwargs):
    if constant.ENABLE_DEBUG:
        print(*args, **kwargs)


def time_filter(data, time_range):
    if time_range[0] is None:
        time_range[0] = -1
    if time_range[1] is None:
        time_range[1] = data[-1]["Time"]
    data = list(filter(lambda x: time_range[0] <= x["Time"] <= time_range[1], data))
    return data


def compose_packet_logs(file_range, pattern=None):
    if pattern is None:
        pattern = "output/packet_log/packet-0.log"
    compose_data = []
    if file_range == "all":
        # Suppose maximum numbers of file is 1000
        file_range = [1000]
    try:
        for idx in range(*file_range):
            with open(pattern.replace("0", str(idx)), 'r') as f:
                for line in f.readlines():
                    compose_data.append(json.loads(line.replace("'", '"')))
    except Exception as e:
        debug_print("Log file ended at {}".format(idx))
    finally:
        return compose_data


def plot_cwnd(log_file, rows=None, trace_file=None, time_range=None, scatter=False, file_range=None, sender=None):
    if not constant.USE_CWND:
        print("Your congestion control don't use windows~")
        return
    plt_data = []
    if file_range:
        plt_data = compose_packet_logs(file_range)
    else:
        with open(log_file, 'r') as f:
            for line in f.readlines():
                plt_data.append(json.loads(line.replace("'", '"')))
    if not sender:
        sender = [1]
    # filter the packet at sender
    plt_data = list(filter(lambda x: x["Type"] == 'S' and x["Position"] == 0 and x["Sender_id"] in sender, plt_data))
    # plt_data = list(filter(lambda x: x["Drop"] == 0, plt_data))
    # filter by the time
    if time_range:
        plt_data = time_filter(plt_data, time_range)
    # plot "rows" counts data
    if isinstance(rows, int):
        plt_data = plt_data[:rows]

    pic_nums = 1
    font_size = 50
    tick_size = 50

    data_time = []
    data_cwnd = []
    data_Ucwnd = []
    last_cwnd = -1
    for item in plt_data:
        if item["Extra"]["Cwnd"] == last_cwnd:
            continue
        last_cwnd = item["Extra"]["Cwnd"]
        data_time.append(item["Time"])
        data_cwnd.append(item["Extra"]["Cwnd"])
        data_Ucwnd.append(item["Waiting_for_ack_nums"])

    pic = plt.figure(figsize=(50, 30*pic_nums))
    # plot cwnd changing
    ax = plt.subplot(pic_nums, 1, 1)
    if scatter:
        ax.scatter(data_time, data_cwnd, label="cwnd", c='g', s=200)
    else:
        ax.plot(data_time, data_cwnd, label="cwnd", c='g', linewidth=5)
    ax.set_ylabel("Packet", fontsize=font_size)
    ax.set_xlabel("Time / s", fontsize=font_size)
    plt.tick_params(labelsize=tick_size)
    # plt.legend(fontsize=font_size)
    # ax.legend(fontsize=font_size)

    # # plot used cwnd changing
    # ax = plt.subplot(pic_nums, 1, 2)
    # ax.scatter(data_time, data_Ucwnd, c='y', label="used_cwnd")
    # ax.set_ylabel("Packet", fontsize=20)
    # ax.set_xlabel("Time (s)", fontsize=20)
    # plt.tick_params(labelsize=20)
    # plt.legend(fontsize=20)
    handles, labels = ax.get_legend_handles_labels()

    # plot bandwith
    if trace_file:
        tmp_ax = plot_trace(data_time, ax, font_size, tick_size, trace_file)
        handles.extend(tmp_ax.get_legend_handles_labels()[0])
        labels.extend(tmp_ax.get_legend_handles_labels()[1])

    plt.legend(handles, labels, fontsize=font_size)
    plt.savefig("output/cwnd_changing.png")


def plot_trace(data_time, ax, font_size, tick_size, trace_file):
    max_time = data_time[-1]
    trace_list = []
    with open(trace_file, "r") as f:
        for line in f.readlines():
            trace_list.append(list(
                map(lambda x: float(x), line.split(","))
            ))

    st = data_time[0]
    ax = ax.twinx()
    ed = -1
    for idx in range(1, len(trace_list)):
        if trace_list[idx][0] < st:
            continue
        if trace_list[idx][0] > max_time:
            ed = idx
            break
        ax.plot([st, trace_list[idx][0]], [trace_list[idx - 1][1] * 10 ** 6 / BYTES_PER_PACKET] * 2, '--',
                linewidth=5, c='b')
        st = trace_list[idx][0]

    if ed == -1 and trace_list[-1][0] < max_time:
        ax.plot([st, max_time], [trace_list[-1][1] * 10 ** 6 / BYTES_PER_PACKET] * 2, '--',
                label="Different Bandwith", linewidth=5, c='b')
    elif ed != -1:
        ax.plot([st, max_time], [trace_list[ed - 1][1] * 10 ** 6 / BYTES_PER_PACKET] * 2, '--',
                label="Different Bandwith", linewidth=5, c='b')

    ax.set_ylabel("Link bandwith (Packet/s)", fontsize=font_size)
    plt.tick_params(labelsize=tick_size)
    # plt.legend(fontsize=font_size)
    return ax


def plot_rate(log_file, rows=None, trace_file=None, time_range=None, scatter=False, file_range=None, sender=None):
    plt_data = []
    if file_range:
        plt_data = compose_packet_logs(file_range)
    else:
        with open(log_file, 'r') as f:
            for line in f.readlines():
                plt_data.append(json.loads(line.replace("'", '"')))
    if not sender:
        sender = [1]
    # filter the packet at receiver
    plt_data = list(
        filter(lambda x: x["Type"] == 'S' and x["Position"] == 0 and x["Drop"] == 0 and x["Sender_id"] in sender,
               plt_data))
    # plt_data = list(filter(lambda x: x["Drop"] == 0, plt_data))
    # filter by the time
    if time_range:
        plt_data = time_filter(plt_data, time_range)
    # plot "rows" counts data
    if isinstance(rows, int):
        plt_data = plt_data[:rows]

    pic_nums = 1
    font_size = 50
    tick_size = 50

    data_time = []
    data_rate = []
    data_inflight = []
    for idx, item in enumerate(plt_data):
        data_time.append(item["Time"])
        data_rate.append(item["Extra"]["Send_rate"])
        data_inflight.append(item["Waiting_for_ack_nums"])

    pic = plt.figure(figsize=(50, 30 * pic_nums))
    # plot cwnd changing
    ax = plt.subplot(pic_nums, 1, 1)
    if scatter:
        # ax.scatter(data_time, data_throughput, label="Throughput", c='g', s=200)
        ax.scatter(data_time, data_rate, label="Rate", c='black', s=200)
        # ax.scatter(data_time, data_inflight, label="Inflight", c='r', s=200)
    else:
        # ax.plot(data_time, data_throughput, label="Throughput", c='g')
        ax.plot(data_time, data_rate, label="Rate", c='black', linewidth=5)
        # ax.plot(data_time, data_inflight, label="Inflight", c='r', linewidth=5)
    ax.set_ylabel("Packet Numbers", fontsize=font_size)
    ax.set_xlabel("Time / s", fontsize=font_size)
    plt.tick_params(labelsize=tick_size)
    handles, labels = ax.get_legend_handles_labels()

    # plot bandwith
    if trace_file:
        tmp_ax = plot_trace(data_time, ax, font_size, tick_size, trace_file)
        handles.extend(tmp_ax.get_legend_handles_labels()[0])
        labels.extend(tmp_ax.get_legend_handles_labels()[1])

    plt.legend(handles, labels, fontsize=font_size)
    plt.savefig("output/rate_changing.png")


def plot_bbr(log_file, rows=None, trace_file=None, time_range=None, scatter=False, file_range=None, sender=None):
    plt_data = []
    if file_range:
        plt_data = compose_packet_logs(file_range)
    else:
        with open(log_file, 'r') as f:
            for line in f.readlines():
                plt_data.append(json.loads(line.replace("'", '"')))
    if not sender:
        sender = [1]
    # filter the packet at receiver
    plt_data = list(filter(lambda x: x["Type"] == 'A' and x["Position"] == 1 and x["Drop"] == 0 and x["Sender_id"] in sender, plt_data))
    # plt_data = list(filter(lambda x: x["Drop"] == 0, plt_data))
    # filter by the time
    if time_range:
        plt_data = time_filter(plt_data, time_range)
    # plot "rows" counts data
    if isinstance(rows, int):
        plt_data = plt_data[:rows]

    pic_nums = 1
    font_size = 50
    tick_size = 50

    data_time = []
    data_throughput = []
    data_bdp = []
    data_inflight = []
    for idx, item in enumerate(plt_data):
        if "delivered" not in item["Extra"]:
            continue
        if idx and item["Time"] < data_time[-1]:
            print("error order!")
        data_time.append(item["Time"])
        # used_time = (item["Time"] - item["Create_time"] - item["Send_delay"] - item["Pacing_delay"])
        used_time = item["Latency"]
        data_throughput.append((idx+1-item["Extra"]["delivered"]) / used_time)
        data_bdp.append(item["Extra"]["max_bw"] * item["Extra"]["min_rtt"] if item["Extra"]["max_bw"] != float("-inf") else 0)
        data_inflight.append(item["Waiting_for_ack_nums"])
    # print(data_time)
    # print(data_throughput)
    # print(data_bdp)

    pic = plt.figure(figsize=(50, 30 * pic_nums))
    # plot cwnd changing
    ax = plt.subplot(pic_nums, 1, 1)
    if scatter:
        # ax.scatter(data_time, data_throughput, label="Throughput", c='g', s=200)
        ax.scatter(data_time, data_bdp, label="BDP", c='black', s=200)
        ax.scatter(data_time, data_inflight, label="Inflight", c='r', s=200)
    else:
        # ax.plot(data_time, data_throughput, label="Throughput", c='g')
        ax.plot(data_time, data_bdp, label="BDP", c='black', linewidth=5)
        ax.plot(data_time, data_inflight, label="Inflight", c='r', linewidth=5)
    ax.set_ylabel("Packet Numbers", fontsize=font_size)
    ax.set_xlabel("Time / s", fontsize=font_size)
    plt.tick_params(labelsize=tick_size)
    handles, labels = ax.get_legend_handles_labels()

    # plot bandwith
    if trace_file:
        tmp_ax = plot_trace(data_time, ax, font_size, tick_size, trace_file)
        handles.extend(tmp_ax.get_legend_handles_labels()[0])
        labels.extend(tmp_ax.get_legend_handles_labels()[1])

    plt.legend(handles, labels, fontsize=font_size)
    plt.savefig("output/throughput_changing.png")


if __name__ == '__main__':

    log_packet_file = "output/packet_log/packet-0.log"
    trace_file = "config/trace.txt"
    new_trace_file = "scripts/first_group/traces_1.txt"
    # analyze_pcc_emulator(log_packet_file, time_range=None, scatter=False, trace_file=new_trace_file, file_range="all")
    # plot_cwnd(log_packet_file, None, trace_file=new_trace_file, time_range=None, scatter=False, file_range="all")
    plot_rate(log_packet_file, file_range="all", scatter=False)


class Packet(object):
    _packet_id = 1

    def __init__(self,
                 create_time,
                 next_hop,
                 offset,
                 payload,
                 packet_id=None,
                 packet_size=1500,
                 packet_type="S",
                 drop=False,
                 send_delay=.0,
                 pacing_delay=.0,
                 latency=.0,
                 block_info={}
                 ):
        self.packet_type = packet_type
        self.create_time = create_time
        self.finish_time = -1
        self.next_hop = next_hop
        self.offset = offset
        self.packet_id = packet_id
        self.payload = payload
        self.packet_size = packet_size
        self.drop = drop

        self.send_delay = send_delay
        self.pacing_delay = pacing_delay
        self.latency = latency
        self.extra = {}
        self.block_info = block_info

        if packet_id is None:
            self.packet_id = Packet._get_next_packet()

    @classmethod
    def _get_next_packet(cls):
        result = cls._packet_id
        cls._packet_id += 1
        return result

    @classmethod
    def create_normal_packet(cls, cur_time, **kwargs):
        return Packet(
                    create_time=cur_time,
                    next_hop=0,
                    offset=-1,
                    packet_size=kwargs["packet_size"],
                    payload=-1)

    def parse(self):

        return [self.packet_type,
                self.next_hop,
                self.latency,
                self.drop,
                self.packet_id]

    def get_rtt(self):
        return self.pacing_delay+self.latency

    def trans2dict(self):
        print_data = {
            "Type": self.packet_type,
            "Position": self.next_hop,
            "Send_delay": self.send_delay,
            "Pacing_delay" : self.pacing_delay,
            "Latency": self.latency,
            "Drop": 1 if self.drop else 0,
            "Packet_id": self.packet_id,
            "Create_time": self.create_time,
            "Offset" : self.offset,
            "Payload" : self.payload,
            "Packet_size" : self.packet_size,
            "Extra" : self.extra,
            "Block_info" : self.block_info
        }
        return print_data

    def create_retrans_packet(self, cur_time):
        """create a new packet for retransmission."""
        return Packet(create_time=cur_time,
                      next_hop=0,
                      offset=self.offset,
                      packet_size=self.packet_size,
                      payload=self.payload,
                      block_info=self.block_info)

    def get_hash_val(self):
        """get the hash value of this packet according to it's member variables."""
        tmp = self.trans2dict()
        MOD = 1234567891
        ret = 0
        for key, val in tmp.items():
            if isinstance(val, dict):
                for k, v in val.items():
                    ret = (ret * (hash(v)%MOD) ) % MOD
            else:
                ret = (ret * (hash(val)%MOD) ) % MOD
        return ret

    def is_miss_ddl(self, cur_time=None):
        # use current time to judge
        if cur_time and self.finish_time != -1:
            return cur_time > self.finish_time
        # for finished packet
        return cur_time-self.block_info["Create_time"] > self.block_info["Deadline"]

    def __lt__(self, other):
        return self.create_time < other.create_time

    def __str__(self):
        print_data = self.trans2dict()
        return str(print_data)


"""
This demo aims to help player running system quickly by using the pypi library DTP-Emualtor https://pypi.org/project/DTP-Emulator/.
"""
from simple_emulator import PccEmulator, CongestionControl

# We provided a simple algorithms about packet selection to help you being familiar with this competition.
# In this example, it will select the packet according to packet's created time first and radio of rest life time to deadline secondly.
from simple_emulator import Packet_selection

# We provided some simple algorithms about congestion control to help you being familiar with this competition.
# Like Reno and an example about reinforcement learning implemented by tensorflow
from simple_emulator import Reno
# Ensuring that you have installed tensorflow before you use it
# from simple_emulator import RL

# We provided some function of plotting to make you analyze result easily in utils.py
from simple_emulator import analyze_pcc_emulator, plot_cwnd, plot_rate


# Your solution should include packet selection and congestion control.
# So, we recommend you to achieve it by inherit the objects we provided and overwritten necessary method.
class MySolution(Packet_selection, Reno):

    def select_packet(self, cur_time, packet_queue):
        """
        The algorithm to select which packet in 'packet_queue' should be sent at time 'cur_time'.
        See more at https://github.com/Azson/DTP-emulator/tree/pcc-emulator#packet_selectionpy.
        """
        return super().select_packet(cur_time, packet_queue)

    def make_decision(self, cur_time):
        """
        The part of algorithm to make congestion control, which will be call when sender need to send pacekt.
        See more at https://github.com/Azson/DTP-emulator/tree/pcc-emulator#congestion_control_algorithmpy.
        """
        return super().make_decision(cur_time)

    def append_input(self, data):
        """
        The part of algorithm to make congestion control, which will be call when sender get an event about acknowledge or lost from reciever.
        See more at https://github.com/Azson/DTP-emulator/tree/pcc-emulator#congestion_control_algorithmpy.
        """
        return super().append_input(data)


# The file path of packets' log
log_packet_file = "output/packet_log/packet-0.log"

# Use the object you created above
my_solution = MySolution()

# Create the emulator using your solution
# Specify USE_CWND to decide whether or not use crowded windows. USE_CWND=True by default.
# Specify ENABLE_LOG to decide whether or not output the log of packets. ENABLE_LOG=True by default.
# You can get more information about parameters at https://github.com/Azson/DTP-emulator/tree/pcc-emulator#constant
emulator = PccEmulator(
    solution=my_solution,
    USE_CWND=True,
    ENABLE_LOG=True
)

# Run the emulator and you can specify the time for the emualtor's running.
# It will run until there is no packet can sent by default.
emulator.run_for_dur()

# print the debug information of links and senders
emulator.print_debug()

# Output the picture of pcc_emulator-analysis.png
# You can get more information from https://github.com/Azson/DTP-emulator/tree/pcc-emulator#pcc_emulator-analysispng.
analyze_pcc_emulator(log_packet_file, file_range="all")

# Output the picture of cwnd_changing.png
# You can get more information from https://github.com/Azson/DTP-emulator/tree/pcc-emulator#cwnd_changingpng
plot_cwnd(log_packet_file, file_range="all")


#!/usr/bin/python
# -*- coding: utf-8 -*-


from objects.pcc_emulator import PccEmulator
import os, sys, inspect
from config.constant import *
import numpy as np
currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(currentdir)
sys.path.insert(0, parentdir)

import json

from player.aitrans_solution import Solution as s1
from player.aitrans_solution2 import Solution as s2

def cal_qoe(x=0.82):
    block_data = []
    urgency = []
    priorities = []
    qoe = 0
    tmp = [3, 2, 1]
    with open("output/block.log", "r") as f:
        for line in f.readlines():
            block_data.append(json.loads(line.replace("'", '"')))
    for block in block_data:
        priority = float(tmp[int(block['Priority'])] / 3)
        priorities.append(priority)
        if block['Miss_ddl'] == 0:
            urgency.append(1)
        else:
            urgency.append(0)
            priorities[-1] *= -1
    for i in range(len(urgency)):
        qoe += x * priorities[i] + (1 - x) * urgency[i]
    return qoe

def cal_distance(block_file, trace_file, x):
    emulator1 = PccEmulator(
        block_file=block_file,
        trace_file=trace_file,
        queue_range=(MIN_QUEUE, MAX_QUEUE),
        solution=s1()
    )
    emulator1.run_for_dur(float("inf"))
    reno_qoe = cal_qoe(x)

    emulator2 = PccEmulator(
        block_file=block_file,
        trace_file=trace_file,
        queue_range=(MIN_QUEUE, MAX_QUEUE),
        solution=s2()
    )
    emulator2.run_for_dur(float("inf"))
    bbr_qoe = cal_qoe(x)
    return abs(reno_qoe - bbr_qoe)



if __name__ == '__main__':

    block_file = "config/block.txt"
    log_file = "output/pcc_emulator.log"
    log_packet_file = "output/packet_log/packet-0.log"

    x = 0
    qoes = {}
    for i in range(1, 100, 3):
        x = i / 100
        arr = []
        for j in range(1, 101):
            trace_file = "scripts/first_group/traces_" + str(j) + ".txt"
            qoe_distance = cal_distance(block_file, trace_file, x)
            arr.append(qoe_distance)
        arr = np.array(arr, float)
        qoes[x] = np.var(arr)
    best_x = max(qoes, key=qoes.get)
    with open("qoemodel/qoe_model.log","w+") as f:
        f.write(str(qoes) + '\n')
        f.write(str(best_x) + " * priority + " + str(1 - best_x) + " * ddl " )





